#include <stdio.h>
#include <stdlib.h>

#include "bst.h"

// Private functions.
void bst_data_walk_inorder(struct bst_node * node, void (*op)(void * data));
void bst_data_walk_preorder(struct bst_node * node, void (*op)(void * data));
void bst_data_walk_postorder(struct bst_node * node, void (*op)(void * data));

void bst_node_walk_postorder(struct bst_node * node, void (*op)(struct bst_node * n));

void * bst_data_search(struct bst * bst, struct bst_node * node, void * data);
void * bst_node_minimum(struct bst_node * node);
void * bst_node_maximum(struct bst_node * node);
void bst_node_free(struct bst_node * node);

struct bst_node * bst_node_search(struct bst * bst, struct bst_node * node, void * data);

struct bst_node * bst_transplant(struct bst * bst, struct bst_node * a, struct bst_node * b);

////////////////////////////////////////////////////////////////////////////////
struct bst *
bst_new(bool (*le)(void * a, void * b),
	bool (*eq)(void * a, void * b))
   {
   if ( (NULL == le) || (NULL == eq) )
      return NULL;

   struct bst * bst = malloc(sizeof(struct bst));
   if (NULL == bst)
      return NULL;

   bst->root = NULL;
   bst->size = 0;
   bst->le = le;
   bst->eq = eq;
   return bst;
   }


////////////////////////////////////////////////////////////////////////////////
void
bst_delete(struct bst * bst)
   {
   if (NULL == bst) return;
   if (0 < bst->size)
      {
      // Walk the tree and delete the nodes.
      // This may cause memory leaks, so hopefully the user has already
      // emptied the tree.
      bst_node_walk_postorder(bst->root, bst_node_free);
      }
   free(bst);
   }

////////////////////////////////////////////////////////////////////////////////
void
bst_node_free(struct bst_node * node)
   {
   free(node);
   }

////////////////////////////////////////////////////////////////////////////////
void *
bst_insert(struct bst * bst, void * data)
   {
   if ( (NULL == bst) || (NULL == data) )
      return NULL;

   struct bst_node * new = malloc(sizeof(struct bst_node));
   if (NULL == new)
      return NULL;
   new->left = new->right = new->parent = NULL;
   new->data = data;

   if (NULL == bst->root)
      {
      bst->root = new;
      return new->data;
      }

   struct bst_node * p = NULL;
   struct bst_node * c = bst->root;

   while (NULL != c)
      {
      p = c;
      if (bst->le(new->data, c->data))
	 c = c->left;
      else
	 c = c->right;
      }
   new->parent = p;
   if (bst->le(new->data, p->data))
      p->left = new;
   else
      p->right = new;

   bst->size++;
   return new->data;
   }


////////////////////////////////////////////////////////////////////////////////
void bst_walk_inorder(struct bst * bst, void (*op)(void * data))
   {
   if ( (NULL == bst) || (NULL == bst->root) || (NULL == op) )
      return;

   bst_data_walk_inorder(bst->root, op);
   }

////////////////////////////////////////////////////////////////////////////////
void bst_walk_postorder(struct bst * bst, void (*op)(void * data))
   {
   if ( (NULL == bst) || (NULL == bst->root) || (NULL == op) )
      return;

   bst_data_walk_postorder(bst->root, op);
   }

////////////////////////////////////////////////////////////////////////////////
void bst_walk_preorder(struct bst * bst, void (*op)(void * data))
   {
   if ( (NULL == bst) || (NULL == bst->root) || (NULL == op) )
      return;

   bst_data_walk_preorder(bst->root, op);
   }

////////////////////////////////////////////////////////////////////////////////
void bst_data_walk_inorder(struct bst_node * node, void (*op)(void * data))
   {
   if (NULL != node)
      {
      bst_data_walk_inorder(node->left, op);
      op(node->data);
      bst_data_walk_inorder(node->right, op);
      }
   }

////////////////////////////////////////////////////////////////////////////////
void bst_data_walk_postorder(struct bst_node * node, void (*op)(void * data))
   {
   if (NULL != node)
      {
      bst_data_walk_postorder(node->left, op);
      bst_data_walk_postorder(node->right, op);
      op(node->data);
      }
   }

////////////////////////////////////////////////////////////////////////////////
void bst_data_walk_preorder(struct bst_node * node, void (*op)(void * data))
   {
   if (NULL != node)
      {
      op(node->data);
      bst_data_walk_preorder(node->left, op);
      bst_data_walk_preorder(node->right, op);
      }
   }

////////////////////////////////////////////////////////////////////////////////
void bst_node_walk_postorder(struct bst_node * node, void (*op)(struct bst_node * n))
   {
   if (NULL != node)
      {
      bst_node_walk_postorder(node->left, op);
      bst_node_walk_postorder(node->right, op);
      op(node);
      }
   }

////////////////////////////////////////////////////////////////////////////////
void * bst_search(struct bst * bst, void * data)
   {
   if ( (NULL == bst) || (NULL == bst->root) || (NULL == data) )
      return NULL;

   return bst_data_search(bst, bst->root, data);
   }

////////////////////////////////////////////////////////////////////////////////
// Recursive
void * bst_data_search(struct bst * bst, struct bst_node * node, void * data)
   {
   if (NULL == node)
      return NULL;

   if (bst->eq(data, node->data))
      return node->data;

   if (bst->le(data, node->data))
      return bst_data_search(bst, node->left, data);
   else
      return bst_data_search(bst, node->right, data);
   }

////////////////////////////////////////////////////////////////////////////////
// Iterative
void * bst_search_iterative(struct bst * bst, void * data)
   {
   if ( (NULL == bst) || (NULL == bst->root) || (NULL == data) )
      return NULL;

   struct bst_node * node = bst->root;
   while (1)
      {
      if (NULL == node)
	 return NULL;
      
      if (bst->eq(data, node->data))
	 return node->data;
      
      if (bst->le(data, node->data))
	 node = node->left;
      else
	 node = node->right;
      }
   }

////////////////////////////////////////////////////////////////////////////////
void * bst_minimum(struct bst * bst)
   {
   if (NULL == bst)
      return NULL;

   return bst_node_minimum(bst->root);
   }

////////////////////////////////////////////////////////////////////////////////
void * bst_node_minimum(struct bst_node * node)
   {
   while ( (NULL != node) && (NULL != node->left) )
      node = node->left;
   if (NULL == node)
      return NULL;
   return node->data;
   }

////////////////////////////////////////////////////////////////////////////////
void * bst_maximum(struct bst * bst)
   {
   if (NULL == bst)
      return NULL;

   return bst_node_maximum(bst->root);
   }

////////////////////////////////////////////////////////////////////////////////
void * bst_node_maximum(struct bst_node * node)
   {
   while ( (NULL != node) && (NULL != node->right) )
      node = node->right;
   if (NULL == node)
      return NULL;
   return node->data;
   }

////////////////////////////////////////////////////////////////////////////////
struct bst_iterator * bst_begin(struct bst * bst)
   {
   if ( (NULL == bst) || (NULL == bst->root) )
      return NULL;

   struct bst_node * min = bst->root;
   while (min->left)
      min = min->left;

   struct bst_iterator * it = malloc(sizeof(struct bst_iterator));
   if (NULL == it)
      return NULL;

   it->bst = bst;
   it->curr = min;
   return it;
   }

////////////////////////////////////////////////////////////////////////////////
struct bst_iterator * bst_end(struct bst * bst)
   {
   if ( (NULL == bst) || (NULL == bst->root) )
      return NULL;

   struct bst_node * max = bst->root;
   while (max->right)
      max = max->right;

   struct bst_iterator * it = malloc(sizeof(struct bst_iterator));
   if (NULL == it)
      return NULL;

   it->bst = bst;
   it->curr = max;
   return it;
   }

////////////////////////////////////////////////////////////////////////////////
struct bst_iterator * bst_next(struct bst_iterator * it)
   {
   if ( (NULL == it) || (NULL == it->bst) || (NULL == it->curr) )
      return NULL;

   if (NULL == it->curr->right)
      {
      struct bst_node * p = it->curr;
      it->curr = it->curr->parent;
      while ( (NULL != it->curr) && (p == it->curr->right) )
	 {
	 p = it->curr;
	 it->curr = it->curr->parent;
	 }

      if (it->curr == NULL)
	 return NULL;
      }
   else
      {
      struct bst_node * min = it->curr->right;
      while (min->left)
	 min = min->left;
      it->curr = min;
      }
   return it;
   }

////////////////////////////////////////////////////////////////////////////////
struct bst_iterator * bst_prev(struct bst_iterator * it)
   {
   if ( (NULL == it) || (NULL == it->bst) || (NULL == it->curr) )
      return NULL;

   if (NULL == it->curr->left)
      {
      struct bst_node * p = it->curr;
      it->curr = it->curr->parent;
      while ( (NULL != it->curr) && (p == it->curr->left) )
	 {
	 p = it->curr;
	 it->curr = it->curr->parent;
	 }

      if (it->curr == NULL)
	 return NULL;
      }
   else
      {
      struct bst_node * max = it->curr->left;
      while (max->right)
	 max = max->right;
      it->curr = max;
      }
   return it;
   }

////////////////////////////////////////////////////////////////////////////////
void * bst_value(struct bst_iterator * it)
   {
   if ( (NULL == it) || (NULL == it->bst) || (NULL == it->curr) )
      return NULL;

   return it->curr->data;
   }

////////////////////////////////////////////////////////////////////////////////
void *
bst_find(struct bst_iterator * it, void * data)
   {
   if ( (NULL == it) || (NULL == it->bst) || (NULL == it->bst->root) || (NULL == data) )
      return NULL;

   return bst_node_search(it->bst, it->bst->root, data);
   }

////////////////////////////////////////////////////////////////////////////////
// Recursive
struct bst_node *
bst_node_search(struct bst * bst, struct bst_node * node, void * data)
   {
   if (NULL == node)
      return NULL;

   if (bst->eq(data, node->data))
      return node;

   if (bst->le(data, node->data))
      return bst_node_search(bst, node->left, data);
   else
      return bst_node_search(bst, node->right, data);
   }

////////////////////////////////////////////////////////////////////////////////
struct bst_node *
bst_transplant(struct bst * bst, struct bst_node * a, struct bst_node * b)
   {
   if ( (NULL == a) || (NULL == b) )
      return NULL;

   if (NULL == a->parent)
      bst->root = b;
   else if (a == a->parent->left)
      a->parent->left = b;
   else
      a->parent->right = b;

   if (NULL != b)
      b->parent = a->parent;

   return a;
   }

////////////////////////////////////////////////////////////////////////////////
void *
bst_remove(struct bst_iterator * it)
   {
   if ( (NULL == it) || (NULL == it->curr) )
      return NULL;

   struct bst_node * y;

   if (NULL == it->curr->left)
      {
      bst_transplant(it->bst, it->curr, it->curr->right);
      y = it->curr->right;
      }
   else if (NULL == it->curr->right)
      {
      bst_transplant(it->bst, it->curr, it->curr->left);
      y = it->curr->left;
      }
   else
      {
      y = bst_node_minimum(it->curr->right);
      if (NULL != y->parent)
	 {
	 bst_transplant(it->bst, y, y->right);
	 y->right = it->curr->right;
	 y->right->parent = y;
	 }
      bst_transplant(it->bst, it->curr, y);
      y->left = it->curr->left;
      y->left->parent = y;
      }

   void * data = it->curr->data;
   free(it->curr);
   it->curr = y;
   return data;
   }

